/* !checksrc! disable COPYRIGHT all */

#include "first.h"

#include "memptr.c"
#include "testutil.c"
#include "testtrace.c"
#include "../../lib/curl_threads.c"
#include "../../lib/curlx/fopen.c"
#include "../../lib/curlx/multibyte.c"
#include "../../lib/curlx/strcopy.c"
#include "../../lib/curlx/strerr.c"
#include "../../lib/curlx/strparse.c"
#include "../../lib/curlx/timediff.c"
#include "../../lib/curlx/timeval.c"
#include "../../lib/curlx/version_win32.c"
#include "../../lib/curlx/wait.c"
#include "../../lib/curlx/warnless.c"
#include "../../lib/curlx/winapi.c"
#include "../../src/toolx/tool_time.c"
#include "cli_ftp_upload.c"
#include "cli_h2_pausing.c"
#include "cli_h2_serverpush.c"
#include "cli_h2_upgrade_extreme.c"
#include "cli_hx_download.c"
#include "cli_hx_upload.c"
#include "cli_tls_session_reuse.c"
#include "cli_upload_pausing.c"
#include "cli_ws_data.c"
#include "cli_ws_pingpong.c"
#include "lib500.c"
#include "lib501.c"
#include "lib502.c"
#include "lib503.c"
#include "lib504.c"
#include "lib505.c"
#include "lib506.c"
#include "lib507.c"
#include "lib508.c"
#include "lib509.c"
#include "lib510.c"
#include "lib511.c"
#include "lib512.c"
#include "lib513.c"
#include "lib514.c"
#include "lib515.c"
#include "lib516.c"
#include "lib517.c"
#include "lib518.c"
#include "lib519.c"
#include "lib520.c"
#include "lib521.c"
#include "lib523.c"
#include "lib524.c"
#include "lib525.c"
#include "lib526.c"
#include "lib530.c"
#include "lib533.c"
#include "lib536.c"
#include "lib537.c"
#include "lib539.c"
#include "lib540.c"
#include "lib541.c"
#include "lib542.c"
#include "lib543.c"
#include "lib544.c"
#include "lib547.c"
#include "lib549.c"
#include "lib552.c"
#include "lib553.c"
#include "lib554.c"
#include "lib555.c"
#include "lib556.c"
#include "lib557.c"
#include "lib558.c"
#include "lib559.c"
#include "lib560.c"
#include "lib562.c"
#include "lib564.c"
#include "lib566.c"
#include "lib567.c"
#include "lib568.c"
#include "lib569.c"
#include "lib570.c"
#include "lib571.c"
#include "lib572.c"
#include "lib573.c"
#include "lib574.c"
#include "lib575.c"
#include "lib576.c"
#include "lib578.c"
#include "lib579.c"
#include "lib582.c"
#include "lib583.c"
#include "lib586.c"
#include "lib589.c"
#include "lib590.c"
#include "lib591.c"
#include "lib597.c"
#include "lib598.c"
#include "lib599.c"
#include "lib643.c"
#include "lib650.c"
#include "lib651.c"
#include "lib652.c"
#include "lib653.c"
#include "lib654.c"
#include "lib655.c"
#include "lib658.c"
#include "lib659.c"
#include "lib661.c"
#include "lib666.c"
#include "lib667.c"
#include "lib668.c"
#include "lib670.c"
#include "lib674.c"
#include "lib676.c"
#include "lib677.c"
#include "lib678.c"
#include "lib694.c"
#include "lib695.c"
#include "lib751.c"
#include "lib753.c"
#include "lib758.c"
#include "lib757.c"
#include "lib766.c"
#include "lib1156.c"
#include "lib1301.c"
#include "lib1308.c"
#include "lib1485.c"
#include "lib1500.c"
#include "lib1501.c"
#include "lib1502.c"
#include "lib1506.c"
#include "lib1507.c"
#include "lib1508.c"
#include "lib1509.c"
#include "lib1510.c"
#include "lib1511.c"
#include "lib1512.c"
#include "lib1513.c"
#include "lib1514.c"
#include "lib1515.c"
#include "lib1517.c"
#include "lib1518.c"
#include "lib1520.c"
#include "lib1522.c"
#include "lib1523.c"
#include "lib1525.c"
#include "lib1526.c"
#include "lib1527.c"
#include "lib1528.c"
#include "lib1529.c"
#include "lib1530.c"
#include "lib1531.c"
#include "lib1532.c"
#include "lib1533.c"
#include "lib1534.c"
#include "lib1535.c"
#include "lib1536.c"
#include "lib1537.c"
#include "lib1538.c"
#include "lib1540.c"
#include "lib1541.c"
#include "lib1542.c"
#include "lib1545.c"
#include "lib1549.c"
#include "lib1550.c"
#include "lib1551.c"
#include "lib1552.c"
#include "lib1553.c"
#include "lib1554.c"
#include "lib1555.c"
#include "lib1556.c"
#include "lib1557.c"
#include "lib1558.c"
#include "lib1559.c"
#include "lib1560.c"
#include "lib1564.c"
#include "lib1565.c"
#include "lib1567.c"
#include "lib1568.c"
#include "lib1569.c"
#include "lib1571.c"
#include "lib1576.c"
#include "lib1582.c"
#include "lib1587.c"
#include "lib1591.c"
#include "lib1592.c"
#include "lib1593.c"
#include "lib1594.c"
#include "lib1597.c"
#include "lib1598.c"
#include "lib1599.c"
#include "lib1662.c"
#include "lib1900.c"
#include "lib1901.c"
#include "lib1902.c"
#include "lib1903.c"
#include "lib1905.c"
#include "lib1906.c"
#include "lib1907.c"
#include "lib1908.c"
#include "lib1910.c"
#include "lib1911.c"
#include "lib1912.c"
#include "lib1913.c"
#include "lib1915.c"
#include "lib1916.c"
#include "lib1918.c"
#include "lib1919.c"
#include "lib1920.c"
#include "lib1933.c"
#include "lib1934.c"
#include "lib1935.c"
#include "lib1936.c"
#include "lib1937.c"
#include "lib1938.c"
#include "lib1939.c"
#include "lib1940.c"
#include "lib1945.c"
#include "lib1947.c"
#include "lib1948.c"
#include "lib1955.c"
#include "lib1956.c"
#include "lib1957.c"
#include "lib1958.c"
#include "lib1959.c"
#include "lib1960.c"
#include "lib1964.c"
#include "lib1965.c"
#include "lib1970.c"
#include "lib1971.c"
#include "lib1972.c"
#include "lib1973.c"
#include "lib1974.c"
#include "lib1975.c"
#include "lib1977.c"
#include "lib1978.c"
#include "lib2023.c"
#include "lib2032.c"
#include "lib2082.c"
#include "lib2301.c"
#include "lib2302.c"
#include "lib2304.c"
#include "lib2306.c"
#include "lib2308.c"
#include "lib2309.c"
#include "lib2402.c"
#include "lib2404.c"
#include "lib2405.c"
#include "lib2502.c"
#include "lib2504.c"
#include "lib2505.c"
#include "lib2506.c"
#include "lib2700.c"
#include "lib3010.c"
#include "lib3025.c"
#include "lib3026.c"
#include "lib3027.c"
#include "lib3033.c"
#include "lib3034.c"
#include "lib3100.c"
#include "lib3101.c"
#include "lib3102.c"
#include "lib3103.c"
#include "lib3104.c"
#include "lib3105.c"
#include "lib3207.c"
#include "lib3208.c"
#include "lib1521.c"

const struct entry_s s_entries[] = {
  {"cli_ftp_upload", test_cli_ftp_upload},
  {"cli_h2_pausing", test_cli_h2_pausing},
  {"cli_h2_serverpush", test_cli_h2_serverpush},
  {"cli_h2_upgrade_extreme", test_cli_h2_upgrade_extreme},
  {"cli_hx_download", test_cli_hx_download},
  {"cli_hx_upload", test_cli_hx_upload},
  {"cli_tls_session_reuse", test_cli_tls_session_reuse},
  {"cli_upload_pausing", test_cli_upload_pausing},
  {"cli_ws_data", test_cli_ws_data},
  {"cli_ws_pingpong", test_cli_ws_pingpong},
  {"lib500", test_lib500},
  {"lib501", test_lib501},
  {"lib502", test_lib502},
  {"lib503", test_lib503},
  {"lib504", test_lib504},
  {"lib505", test_lib505},
  {"lib506", test_lib506},
  {"lib507", test_lib507},
  {"lib508", test_lib508},
  {"lib509", test_lib509},
  {"lib510", test_lib510},
  {"lib511", test_lib511},
  {"lib512", test_lib512},
  {"lib513", test_lib513},
  {"lib514", test_lib514},
  {"lib515", test_lib515},
  {"lib516", test_lib516},
  {"lib517", test_lib517},
  {"lib518", test_lib518},
  {"lib519", test_lib519},
  {"lib520", test_lib520},
  {"lib521", test_lib521},
  {"lib523", test_lib523},
  {"lib524", test_lib524},
  {"lib525", test_lib525},
  {"lib526", test_lib526},
  {"lib530", test_lib530},
  {"lib533", test_lib533},
  {"lib536", test_lib536},
  {"lib537", test_lib537},
  {"lib539", test_lib539},
  {"lib540", test_lib540},
  {"lib541", test_lib541},
  {"lib542", test_lib542},
  {"lib543", test_lib543},
  {"lib544", test_lib544},
  {"lib547", test_lib547},
  {"lib549", test_lib549},
  {"lib552", test_lib552},
  {"lib553", test_lib553},
  {"lib554", test_lib554},
  {"lib555", test_lib555},
  {"lib556", test_lib556},
  {"lib557", test_lib557},
  {"lib558", test_lib558},
  {"lib559", test_lib559},
  {"lib560", test_lib560},
  {"lib562", test_lib562},
  {"lib564", test_lib564},
  {"lib566", test_lib566},
  {"lib567", test_lib567},
  {"lib568", test_lib568},
  {"lib569", test_lib569},
  {"lib570", test_lib570},
  {"lib571", test_lib571},
  {"lib572", test_lib572},
  {"lib573", test_lib573},
  {"lib574", test_lib574},
  {"lib575", test_lib575},
  {"lib576", test_lib576},
  {"lib578", test_lib578},
  {"lib579", test_lib579},
  {"lib582", test_lib582},
  {"lib583", test_lib583},
  {"lib586", test_lib586},
  {"lib589", test_lib589},
  {"lib590", test_lib590},
  {"lib591", test_lib591},
  {"lib597", test_lib597},
  {"lib598", test_lib598},
  {"lib599", test_lib599},
  {"lib643", test_lib643},
  {"lib650", test_lib650},
  {"lib651", test_lib651},
  {"lib652", test_lib652},
  {"lib653", test_lib653},
  {"lib654", test_lib654},
  {"lib655", test_lib655},
  {"lib658", test_lib658},
  {"lib659", test_lib659},
  {"lib661", test_lib661},
  {"lib666", test_lib666},
  {"lib667", test_lib667},
  {"lib668", test_lib668},
  {"lib670", test_lib670},
  {"lib674", test_lib674},
  {"lib676", test_lib676},
  {"lib677", test_lib677},
  {"lib678", test_lib678},
  {"lib694", test_lib694},
  {"lib695", test_lib695},
  {"lib751", test_lib751},
  {"lib753", test_lib753},
  {"lib758", test_lib758},
  {"lib757", test_lib757},
  {"lib766", test_lib766},
  {"lib1156", test_lib1156},
  {"lib1301", test_lib1301},
  {"lib1308", test_lib1308},
  {"lib1485", test_lib1485},
  {"lib1500", test_lib1500},
  {"lib1501", test_lib1501},
  {"lib1502", test_lib1502},
  {"lib1506", test_lib1506},
  {"lib1507", test_lib1507},
  {"lib1508", test_lib1508},
  {"lib1509", test_lib1509},
  {"lib1510", test_lib1510},
  {"lib1511", test_lib1511},
  {"lib1512", test_lib1512},
  {"lib1513", test_lib1513},
  {"lib1514", test_lib1514},
  {"lib1515", test_lib1515},
  {"lib1517", test_lib1517},
  {"lib1518", test_lib1518},
  {"lib1520", test_lib1520},
  {"lib1522", test_lib1522},
  {"lib1523", test_lib1523},
  {"lib1525", test_lib1525},
  {"lib1526", test_lib1526},
  {"lib1527", test_lib1527},
  {"lib1528", test_lib1528},
  {"lib1529", test_lib1529},
  {"lib1530", test_lib1530},
  {"lib1531", test_lib1531},
  {"lib1532", test_lib1532},
  {"lib1533", test_lib1533},
  {"lib1534", test_lib1534},
  {"lib1535", test_lib1535},
  {"lib1536", test_lib1536},
  {"lib1537", test_lib1537},
  {"lib1538", test_lib1538},
  {"lib1540", test_lib1540},
  {"lib1541", test_lib1541},
  {"lib1542", test_lib1542},
  {"lib1545", test_lib1545},
  {"lib1549", test_lib1549},
  {"lib1550", test_lib1550},
  {"lib1551", test_lib1551},
  {"lib1552", test_lib1552},
  {"lib1553", test_lib1553},
  {"lib1554", test_lib1554},
  {"lib1555", test_lib1555},
  {"lib1556", test_lib1556},
  {"lib1557", test_lib1557},
  {"lib1558", test_lib1558},
  {"lib1559", test_lib1559},
  {"lib1560", test_lib1560},
  {"lib1564", test_lib1564},
  {"lib1565", test_lib1565},
  {"lib1567", test_lib1567},
  {"lib1568", test_lib1568},
  {"lib1569", test_lib1569},
  {"lib1571", test_lib1571},
  {"lib1576", test_lib1576},
  {"lib1582", test_lib1582},
  {"lib1587", test_lib1587},
  {"lib1591", test_lib1591},
  {"lib1592", test_lib1592},
  {"lib1593", test_lib1593},
  {"lib1594", test_lib1594},
  {"lib1597", test_lib1597},
  {"lib1598", test_lib1598},
  {"lib1599", test_lib1599},
  {"lib1662", test_lib1662},
  {"lib1900", test_lib1900},
  {"lib1901", test_lib1901},
  {"lib1902", test_lib1902},
  {"lib1903", test_lib1903},
  {"lib1905", test_lib1905},
  {"lib1906", test_lib1906},
  {"lib1907", test_lib1907},
  {"lib1908", test_lib1908},
  {"lib1910", test_lib1910},
  {"lib1911", test_lib1911},
  {"lib1912", test_lib1912},
  {"lib1913", test_lib1913},
  {"lib1915", test_lib1915},
  {"lib1916", test_lib1916},
  {"lib1918", test_lib1918},
  {"lib1919", test_lib1919},
  {"lib1920", test_lib1920},
  {"lib1933", test_lib1933},
  {"lib1934", test_lib1934},
  {"lib1935", test_lib1935},
  {"lib1936", test_lib1936},
  {"lib1937", test_lib1937},
  {"lib1938", test_lib1938},
  {"lib1939", test_lib1939},
  {"lib1940", test_lib1940},
  {"lib1945", test_lib1945},
  {"lib1947", test_lib1947},
  {"lib1948", test_lib1948},
  {"lib1955", test_lib1955},
  {"lib1956", test_lib1956},
  {"lib1957", test_lib1957},
  {"lib1958", test_lib1958},
  {"lib1959", test_lib1959},
  {"lib1960", test_lib1960},
  {"lib1964", test_lib1964},
  {"lib1965", test_lib1965},
  {"lib1970", test_lib1970},
  {"lib1971", test_lib1971},
  {"lib1972", test_lib1972},
  {"lib1973", test_lib1973},
  {"lib1974", test_lib1974},
  {"lib1975", test_lib1975},
  {"lib1977", test_lib1977},
  {"lib1978", test_lib1978},
  {"lib2023", test_lib2023},
  {"lib2032", test_lib2032},
  {"lib2082", test_lib2082},
  {"lib2301", test_lib2301},
  {"lib2302", test_lib2302},
  {"lib2304", test_lib2304},
  {"lib2306", test_lib2306},
  {"lib2308", test_lib2308},
  {"lib2309", test_lib2309},
  {"lib2402", test_lib2402},
  {"lib2404", test_lib2404},
  {"lib2405", test_lib2405},
  {"lib2502", test_lib2502},
  {"lib2504", test_lib2504},
  {"lib2505", test_lib2505},
  {"lib2506", test_lib2506},
  {"lib2700", test_lib2700},
  {"lib3010", test_lib3010},
  {"lib3025", test_lib3025},
  {"lib3026", test_lib3026},
  {"lib3027", test_lib3027},
  {"lib3033", test_lib3033},
  {"lib3034", test_lib3034},
  {"lib3100", test_lib3100},
  {"lib3101", test_lib3101},
  {"lib3102", test_lib3102},
  {"lib3103", test_lib3103},
  {"lib3104", test_lib3104},
  {"lib3105", test_lib3105},
  {"lib3207", test_lib3207},
  {"lib3208", test_lib3208},
  {"lib1521", test_lib1521},
  {NULL, NULL}
};

#include "first.c"
