/*****************************************************************************/
/*                                  _   _ ____  _                            */
/*  Project                     ___| | | |  _ \| |                           */
/*                             / __| | | | |_) | |                           */
/*                            | (__| |_| |  _ <| |___                        */
/*                             \___|\___/|_| \_\_____|                       */
/*                                                                           */
/* Copyright (C) Daniel Stenberg, <daniel@haxx.se>, et al.                   */
/*                                                                           */
/* This software is licensed as described in the file COPYING, which         */
/* you should have received as part of this distribution. The terms          */
/* are also available at https://curl.se/docs/copyright.html.                */
/*                                                                           */
/* You may opt to use, copy, modify, merge, publish, distribute and/or sell  */
/* copies of the Software, and permit persons to whom the Software is        */
/* furnished to do so, under the terms of the COPYING file.                  */
/*                                                                           */
/* This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY */
/* KIND, either express or implied.                                          */
/*                                                                           */
/* SPDX-License-Identifier: curl                                             */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/

/*      Use program CURLCL as interface to the curl command line tool        */

             CMD        PROMPT('File transfer utility')

             PARM       KWD(CMDARGS) TYPE(*CHAR) LEN(5000) VARY(*YES *INT2)   +
                          CASE(*MIXED) EXPR(*YES) MIN(1)                      +
                          PROMPT('curl command arguments')
